<?php

/*
 * LICENSE
 * 1. If you like to use this class for personal purposes, it's free.
 * 2. For comercial purposes, please contact me. I'll send a license to you.
 * 3. When you copy the framework you must copy this notice with the source
 *    code. You may alter the source code, but you have to put the original with
 *    your altered version.
 * 4. The license is for all files included in this bundle.
 *
 * HISTORY
 * - 2003-11-03
 *   - Fixed tiny "Notice" bug in export().
 * - 2003-10-29
 *   - sort() allows to use SORT_NULL in combination with any other flag, e.g.
 *     sort("a SORT_NAT SORT_NULL b SORT_NUMERIC SORT_NULL").
 * - 2003-10-28
 *   - Bugfix: update() didn't handled numeric field names properly.
 *   - Speeded up insert() and update() a bit.
 * - 2003-10-27
 *   - Works in both PHP <=4.3.1 and >=4.3.2 now (fgetcsv() was changed).
 *   - Added parameter MyCSV(..., length), defaults to 10k instead 100k now.
 *   - Bugfix: drop_field("id") was allowed by mistake, returns false now.
 * - 2003-10-26
 *   - flock() in MyCSV() doen't block/kill the web server any more when trying
 *     to lock the file twice. Now it triggers an user warning.
 *   - Added close() but made it private because in general it's not needed.
 *   - Added tablename().
 *   - Improved the file reader a lot. ID column may be anywhere in the file or
 *     may be completely missing. MyCSV() adds missing IDs immediatelly.
 *   - Bugfix: write() improved, IDs may contain double quotes now.
 *   - add_field() returns false if field name contains any special character.
 *   - Renamed property _tablename into filename.
 * - 2003-10-20
 *   - Added second parameter to ad_field(..., $afterField).
 *   - Added first() and last().
 *   - Bugfix: delete(0) didn't work, it killed the whole table.
 *   - Bugfix: sort("id ...") works (calls ksort() or krsort() internaly).
 *   - Bugfix: sort("... desc") didn't work because it was case sensitive.
 *   - A bunch of small fixes.
 * - 2003-07-16
 *   - Added SORT_NAT (natural sort) to sort().
 *   - SORT_NULL handles "0" as empty too.
 * - 2003-06-02
 *   - Added some possible field delimiters (":", "\0" and space).
 * - 2003-05-27
 *   - Bugfix: Multiple calls of sort() are working now.
 *   - Bugfix: join() resets the internal array pointer now.
 * - 2003-05-16
 *   - Added: ksort()/krsort(). Similar to sort("id ...")/sort("id DESC ...").
 *   - Bugfix: sort() don't mix up the IDs any more (bases on uasort() now).
 *   - Added sorting type SORT_NULL to sort(). This always moves empty elements
 *     to the end of the table.
 *   - Bugfix: Files can be write-protected, then write() triggers a warning.
 *   - Bugfix: rand() can return more than one random IDs now, e.g. rand(5).
 * - 2003-04-03
 *   - Improved data security a lot. File locking keeps the file locked until
 *     the script ended. No file sharing problems possible any more.
 *   - Added ids(), min(), max() and rand(). They all return one or more ID
 *     numbers. Fetch the corresponding data rows using $table->data($id).
 *   - Added filesystem functions exists() and is_writeable().
 *   - Added some missing methods and aliases, e.g. count(), fetch_assoc(), ...
 *   - No need to use add_field() any more, initial insert() will handle this.
 *   - Ordered the methods in logical groups: SQL/Array/Filesystem functions.
 * - 2003-04-01
 *   - Bugfix: update() uses array_merge now. This allows particular updates.
 *   - Added: id_exists() checks if an ID exists (see array_key_exists()).
 *   - Added: ids() returns all ID numbers (see array_keys()).
 *   - Added: rand() returns a random ID number (see array_rand()).
 * - 2003-03-16
 *   - File rows will be separated by "\r\n" now. Compatibility to Windows.
 *   - Added export() which returns the CSV file as a string.
 *   - Bugfix: insert_id() works now.
 *   - A lot of small changes. Works with E_ALL on. Faster too.
 * - 2003-01-20
 *   - drop_field() works now.
 * - 2002-11-21
 *   - Variable delimiter added, ',' and ';' are supported now.
 *
 * KNOWN BUGS/TODO
 * - sort("... id") don't works properly in any case!
 * - fetch_array() and fetch_row() are not supported, use fetch_assoc() or
 *   each() instead.
 * - num_fields(), list_fields() etc. are not supported, use
 *   count($table->fields) etc. instead.
 * - create() or create_table() is not supported, use add_field() instead.
 * - What about some kind of GROUP BY?
 * - Add where($filter = ""), e.g.
 *   where("field='n' OR b <= 8 AND strtolower(substr(r,0,1))='x'") becomes
 *   eval("$data['field']=='n' OR $data['b'] <= 8 AND strtolower(substr($data['r'],0,1))=='x'");
 * - Add limit(15, 2) / limit(2) / limit("15,2").
 */

/**
 * Another sorting type flag for use in MyCSV::sort().
 */
define('SORT_NAT', 16);

/**
 * Another sorting type flag for use in MyCSV::sort().
 */
define('SORT_NULL', 32);

/**
 * TM::MyCSV, a text file based database complement.
 *
 * This class handles standard CSV or TXT text files like database tables. It
 * supports most benefits of both SQL tables and PHP arrays. It doesn't need
 * a real database management system nor does it require any knowlege of the
 * SQL language. It hides any filesystem functions so you don't have to deal
 * with file pointers, field delimiters, escape sequences and so on. Because
 * it uses the widespreaded standard CSV file format you are able to create,
 * read and update the tables using any spreadsheet software. It supports user
 * definied table sort similar to ORDER BY, auto incremented ID numbers and
 * lots more.
 *
 * See {@link MyCSV()} or {@link dump()} for some examples.
 *
 * @author Thiemo Mttig (http://maettig.com/)
 * @version 2003-10-29
 */
class MyCSV
{
    /**
     * Array containing all the table field names. First have to be "id".
     *
     * @var string $fields
     */
    var $fields = array("id");
  
    /**
     * Two dimensional associative array containing all the table row data.
     *
     * @var array $data
     */
    var $data = array();
  
    /**
     * The field delimiter for separating values in the CSV file. Default is ","
     * (default CSV style). If not, the class tries to use ";" (European/German
     * CSV style), "\t", "\0", "|", "&" (URI encoded/parameter style), ":" (Unix
     * /etc/passwd style) and " " (log file style). Normaly you don't have to
     * touch this variable. Simply choose your delimiter when creating your
     * initial CSV file.
     *
     * @var string $delimiter Field delimiter.
     */
    var $delimiter = ",";
  
    /**
     * @var int $insert_id Last insert ID.
     * @see insert()
     */
    var $insert_id = 0;
  
    /**
     * File name of the CSV table with or without the .csv file name extension.
     * Don't change this cause write() will not realize you want another file.
     *
     * @var string $filename File name of the .csv file.
     * @access private
     */
    var $filename;
    
    /**
     * @var bool $_fp Ressource handle to the CSV file already opened.
     * @access private
     */
    var $_fp = false;
  
    /**
     * @var bool $_bug4_3_2 Flag, cause fgetcsv() changed in PHP 4.3.2.
     * @access private
     */
    var $_is4_3_2 = false;
  
    /**
     * Reads a CSV file and returns it as a MyCSV object.
     *
     * Reads a table into a new MyCSV object. The file name may be entered with
     * or without the <code>.csv</code> file extension. If the file does not
     * exist it will be created when calling {@link write()}. Set <i>length</i>
     * to the maximum number of bytes per row you expect (as you did in
     * fgetcsv()). Default is 10000 bytes per line. Setting this to 1000 may
     * speed up the method up to 20 times if you'r sure there is no longer line.
     *
     * For example, create a file called <code>table.csv</code> with the
     * following content and call the script below.
     *
     * <pre>id,value
     * 3,Example
     * 4,Another value
     * 7,Blue</pre>
     *
     * <pre><?php
     * require_once("MyCSV.class.php");
     * $table = new MyCSV("table");
     * while ($row = $table->each()) {
     *     echo $row['id'] . " is " . $row['value'] . "<br>";
     * }
     * ?></pre>
     *
     * @param tablename string
     * @param length int
     * @return MyCSV
     */
    function MyCSV($tablename, $length = 10000)
    {
        // Check the version because fgetcsv() becomes incompatible in PHP 4.3.2.
        if (substr(PHP_VERSION, 0, 5) >= "4.3.2") $this->_is4_3_2 = true;
      
        $this->filename = $tablename;
        // Add default file extension if missing.
        if (! preg_match('/\.\w+$/', $this->filename)) $this->filename .= ".csv";
      
        // Break if the CSV file for this table does not exist.
        if (! file_exists($this->filename)) return false;
      
        // Open the CSV file for exclusive reading and writing OR reading only.
        if (is_writeable($this->filename)) $mode = "r+b"; else $mode = "rb";
        $this->_fp = fopen($this->filename, $mode);
        if (! $this->_fp) return false;
      
        // Try to lock the file but give up after some failures.
        //- flock($this->_fp, LOCK_EX);
        for ($tries = 5; $tries-- > 0; sleep(1))
        {
            if (flock($this->_fp, LOCK_EX + LOCK_NB)) break;
            if ($tries) continue;
            // Triggers an user error if the CSV file is flocked already.
            user_error("MyCSV::MyCSV() failed ($this->filename is locked already)", E_USER_WARNING);
            return false;
        }
      
        // Try some delimiters, but use "," at least if nothing was found.
        $delimiters = ",;\t\0|&: ,";
        while (count($this->fields) < 2)
        {
            $this->delimiter = $delimiters[0];
            $delimiters = substr($delimiters, 1);
            if (! $delimiters) break;
            rewind($this->_fp);
            $this->fields = fgetcsv($this->_fp, $length, $this->delimiter);
        }
        // On what position is the ID field? Returns $i = -1 if not found.
        for ($i = count($this->fields) - 1; $i > -1; $i--)
        {
            if (strtolower($this->fields[$i]) == "id") break;
        }
        $lastId = 0;
        $fieldsCount = count($this->fields);
        while ($row = fgetcsv($this->_fp, $length, $this->delimiter))
        {
            // Add missing id numbers.
            $id = isset($row[$i]) ? $row[$i] : $lastId + 1;
            $lastId = max($id, $lastId);
            $count = min($fieldsCount, count($row));
            for ($c = 0; $c < $count; $c++)
            {
                // Because fgetcsv() becomes incompatible in PHP 4.3.2.
                if ($this->_is4_3_2) $row[$c] = stripslashes($row[$c]);
                $this->data[$id][$this->fields[$c]] = $row[$c];
            }
        }
        // Always move the id column to the front.
        unset($this->fields[$i]);
        array_unshift($this->fields, "id");
      
        return true;
    }
  
    //--------------------------------------------------------------------------
  
    // PHP MySQL and SQL functions.
  
    /**
     * Adds a new field (column) to the table. Returns false on failure, e.g.
     * if the field already exists.
     *
     * @param field string
     * @param afterField string
     * @return bool
     * @see insert(), drop_field()
     */
    function add_field($field, $afterField = "")
    {
        // Break if the field name contains invalid characters or already exists.
        if (! preg_match('/^\w+$/is', $field) || in_array($field, $this->fields))
        {
            return false;
        }
        if ($afterField)
        {
            $newFields = array();
            foreach ($this->fields as $oldField)
            {
                $newFields[] = $oldField;
                if ($oldField == $afterField) $newFields[] = $field;
            }
            $this->fields = $newFields;
        }
        else $this->fields[] = $field;
        return true;
    }
  
    /**
     * Moves the internal row pointer to the specified row number. This is an
     * alias for {@link seek()}.
     *
     * @param row_number int
     * @return bool
     */
    function data_seek($row_number)
    {
        return $this->seek($row_number);
    }
  
    /**
     * Deletes a table row specified by the <i>id</i>. Deletes all rows if no
     * <i>id</i> is given.
     *
     * @param id int
     * @return void
     */
    function delete($id = null)
    {
        // Delete one row if an id is given.
        if (isset($id)) unset($this->data[$id]);
        // Delete all rows if no id was given (or id is null).
        else $this->data = array();
    }
  
    /**
     * Deletes a field/column from the table.
     *
     * @param field string
     * @return bool
     */
    function drop_field($field)
    {
        if (strtolower($field) == "id") return false;
        while (list($id) = each($this->data))
        {
            unset($this->data[$id][$field]);
        }
        array_splice($this->fields, array_search($field, $this->fields), 1);
        return true;
    }

    /**
     * Clears the table. Remove all columns and all fields too.
     *
     * @return void
     */
    function drop_table()
    {
        $this->fields = array("id");
        $this->data = array();
    }

    /**
     * Gets the current data row and increase the internal pointer. This is an
     * alias for {@link each()}.
     *
     * @return array
     */
    function fetch_assoc()
    {
        return $this->each();
    }

    /**
     * Imports a MySQL table.
     *
     * <i>Warning! This method is EXPERIMENTAL! The behaviour of this method may
     * change in a future release of this class.</i>
     *
     * Any data already in the object will be lost. Returns false on failure.
     *
     * @param source mixed
     * @return void
     */
    function fromMySQL($source = "")
    {
        // Assume source is a string containing the table name.
        if (is_string($source))
        {
            // Use the internal table name already set in the constructor, hopefully.
            if (! $source) list($source) = explode(".", basename($this->filename));
            $source = mysql_query("SELECT * FROM " . $source);
        }
        // Break if the source is still false.
        if (! $source) return false;
        $this->fields = array("id");
        $this->data = array();
        while ($row = mysql_fetch_assoc($source)) $this->insert($row);
    }
  
    /**
     * Inserts a new table row using the next free auto incremented ID number.
     *
     * @param data array
     * @return void
     */
    function insert(&$data)
    {
        // First auto increment id is always 1.
        if (! $this->data) $this->insert_id = 1;
        else $this->insert_id = max(array_keys($this->data)) + 1;
        // If data contains an unused id number, use it.
        if (isset($data['id']) && strlen($data['id']) && ! isset($this->data[$data['id']]))
        {
            $this->insert_id = $data['id'];
        }
       
        $this->data[$this->insert_id] = $data;
       
        // Fetch missing field/column names from the first data row if needed.
        // This can be used instead of add_field().
        if (! isset($this->fields) || count($this->fields) <= 1)
        {
            $this->fields = $data;
            unset($this->fields['id']);
            $this->fields = array_merge(array("id"), array_keys($this->fields));
        }
    }
  
    /**
     * Gets the ID generated by the previous insert() call.
     *
     * @return int
     */
    function insert_id()
    {
        return $this->insert_id;
    }
  
    /**
     * Performs an inner join with another table on a foreign key. This adds
     * temporary columns to the table.
     *
     * @param rightTable array
     * @param foreignKey string
     * @return void
     */
    function join(&$rightTable, $foreignKey)
    {
        reset($this->data);

        while (list($id) = each($this->data))
        {
            if (strtolower($foreignKey) == "id") $fid = $id;
            else $fid = $this->data[$id][$foreignKey];
            if (isset($rightTable->data[$fid]))
            {
                $this->data[$id] += (array)$rightTable->data[$fid];
            }
        }

        // Reset the internal pointer.
        reset($this->data);
    }
  
    /**
     * Gets the number of rows in the table.
     *
     * @return int
     */
    function num_rows()
    {
        return count($this->data);
    }
  
    /**
     * Gets the table name without the default ".csv" file extension.
     *
     * @return string
     */
    function tablename()
    {
        return preg_replace('{^\./|\.csv$}', '', $this->filename);
    }

    /**
     * Updates a table row with some new field/value pairs.
     *
     * Examples:
     *
     * <pre>$table->update(array(...), 3);
     * $table->update(array('id' => 3, ...));
     * $table->update(array('id' => 7, ...), 3); // Moves ID 3 to ID 7</pre>
     *
     * @param data array
     * @param id int
     * @return bool
     */
    function update(&$data, $id = null)
    {
        // update(array(...)) without any ID doesn't make sense.
        if (! isset($data['id']) && ! isset($id)) return false;
        // update(array(...), 3) becomes update(array('id' => 3, ...), 3)
        if (! isset($data['id'])) $data['id'] = $id;
        // update(array('id' => 7, ...)) becomes update(array('id' => 7, ...), 7)
        elseif (! isset($id)) $id = $data['id'];
        // update(array('id' => 7, ...), 3) if forbidden if ID 7 already exists.
        elseif ($data['id'] != $id && isset($this->data[$data['id']])) return false;

        //- $this->data[$data['id']] = array_merge($this->data[$id], $data);
        $this->data[$data['id']] = $data + $this->data[$id];

        // update(array('id' => 7, ...), 3) moves ID 3 to 7, so ID 3 is killed.
        if ($data['id'] != $id) unset($this->data[$id]);
        return true;
    }

    //--------------------------------------------------------------------------

    // PHP array functions.

    /**
     * Gets the number of rows in the table. This is an alias for
     * {@link num_rows()}.
     *
     * @return int
     */
    function count()
    {
        return count($this->data);
    }

    /**
     * Gets the current data row and increases the internal pointer. See
     * {@link MyCSV()} for an example.
     *
     * @return array
     */
    function each()
    {
        if (! list($id, $data) = each($this->data)) return false;
        return array('id' => $id) + $data;
    }
  
    /**
     * Sets the internal pointer to the last data row.
     *
     * @return void
     * @see reset()
     */
    function end()
    {
        end($this->data);
    }
  
    /**
     * Checks if the data row specified by the ID exists.
     *
     * @param id int
     * @return bool
     * @see row_exists()
     */
    function id_exists($id)
    {
        return isset($this->data[$id]);
    }
  
    /**
     * Gets an array containing all the IDs of the table.
     *
     * @return array
     * @see min(), max(), first(), last(), prev(), next(), rand()
     */
    function ids()
    {
        return array_keys($this->data);
    }

    /**
     * Sorts the table rows by ID. This is identical to
     * <code>{@link sort}("id")</code> but a little faster.
     *
     * @param sort_flags int
     * @return void
     */
    function ksort($sort_flags = 0)
    {
        ksort($this->data, $sort_flags);
    }

    /**
     * Sorts the table rows by ID in reverse order. This is identical to
     * <code>{@link sort}("id DESC")</code> but a little faster.
     *
     * @param sort_flags int
     * @return void
     */
    function krsort($sort_flags = 0)
    {
        krsort($this->data, $sort_flags);
    }

    /**
     * Gets the smallest ID number used in the table. Typically, this is 1.
     *
     * @return int
     */
    function min()
    {
        return min(array_keys($this->data));
    }

    /**
     * Gets the biggest ID number used in the table. This is often the same as
     * {@link insert_id()} which returns the last inserted ID. But unlike that,
     * max() doesn't depends on a previous call of {@link insert()}.
     *
     * @return int
     */
    function max()
    {
        return max(array_keys($this->data));
    }

    /**
     * Gets the first ID number from the table. This depends on how's the table
     * sorted and isn't identical to {@link min()} in any case.
     *
     * @return int
     * @see last(), prev()
     */
    function first()
    {
        $ids = array_keys($this->data);
        return $ids[0];
    }

    /**
     * Gets the last ID number used in the table. This depends on how's the
     * table sorted and isn't identical to {@link max()} in any case.
     *
     * @return int
     * @see first(), next()
     */
    function last()
    {
        $ids = array_keys($this->data);
        return $ids[count($ids) - 1];
    }

    /**
     * Gets the previous ID number. This is an alias for {@link next()} with
     * <i>offset</i> defaults to -1. Returns false if there is no row at this
     * position.
     *
     * @param int $offset
     * @return int
     * @see next(), first()
     */
    function prev($id, $offset = -1)
    {
        return $this->next($id, $offset);
    }
  
    /**
     * Gets the next ID number. Use <i>offset</i> to get another ID near to the
     * row specified by <i>id</i>. Default is 1 (one forward). Returns false if
     * there is no row at this position.
     *
     * @param int $offset
     * @return int
     * @see prev(), last()
     */
    function next($id, $offset = 1)
    {
        $ids = array_keys($this->data);
        //- Add sort(ids) to return the nearest smaller/bigger ID numbers.
        $i = array_search($id, $ids) + $offset;
        if (! isset($ids[$i])) return false;
        return $ids[$i];
    }

    /**
     * Picks one or more random ID numbers out of the table.
     *
     * @param num_req int
     * @return int
     */
    function rand($num_req = 1)
    {
        return array_rand($this->data, $num_req);
    }
  
    /**
     * Sets the internal pointer to the first data row.
     *
     * @return void
     * @see end(), each()
     */
    function reset()
    {
        reset($this->data);
    }
  
    /**
     * Looks if a data row is already in the table.
     *
     * @param search array
     * @return bool
     * @see id_exists()
     */
    function row_exists(&$search)
    {
        foreach ($this->data as $row)
        {
            reset($search);
            while (list($key, $value) = each($search))
            {
                if (! isset($row[$key]) || $row[$key] != $value) continue 2;
            }
            return true;
        }
        return false;
    }
  
    /**
     * Orders the table rows by one or more columns.
     *
     * Sorting order flags:
     * - ASC or SORT_ASC - Sort in ascending order (default).
     * - DESC or SORT_DESC - Sort in descending order.
     *
     * Sorting type flags:
     * - SORT_REGULAR - Compare items normally (default).
     * - SORT_NUMERIC - Compare items numerically.
     * - SORT_STRING - Compare items as strings.
     * - SORT_NAT - Compare items using a "natural order" algorithm.
     *
     * Special condition flag: SORT_NULL - Move empty elements to the end.
     *
     * No two sorting flags of the same type can be specified after each field.
     * Some examples:
     *
     * <pre>$table->sort("a, b DESC");
     * $table->sort("a b DESC"); // Same as above
     * $table->sort("a", "b", SORT_DESC); // Same as above
     * $table->sort("a SORT_STRING SORT_NULL b SORT_NULL");
     * $table->sort("a SORT_NAT, b SORT_NAT, c");</pre>
     *
     * @param sort_flags mixed
     * @return void
     */
    function sort($sort_flags)
    {
        global $_cmpFields;

        // MyCSV::sort() can be called using array_multisort()-like multiple
        // parameters or a SQL-like string argument.
        if (func_num_args() > 1) $sort_flags = func_get_args();
        else $sort_flags = preg_split('/[,\s]+/', trim($sort_flags));
        // trim(..., ", \t\n\r;") would be better but works in PHP 4.1.0+ only.
    
        // Reset the $GLOBALS['_cmpFields'] array first.
        $_cmpFields = array();
    
        // Calculate the $GLOBALS['_cmpFields'] array for use in MyCSV::_cmp().
        foreach ($sort_flags as $f)
        {
            $f = preg_replace('/^(A|DE)SC$/i', 'SORT_\0', $f);
            // Always use the integer values of predefined constants if available.
            if (defined(strtoupper($f))) $f = constant(strtoupper($f));
        
            // Ignore ascending order but store anything else in the associative array.
            if ($f == SORT_ASC) continue;
            elseif ($f == SORT_DESC) $_cmpFields[count($_cmpFields) - 1]['order'] = -1;
            elseif (is_int($f)) $_cmpFields[count($_cmpFields) - 1]['type'] |= $f;
            else $_cmpFields[] = array('field' => $f, 'order' => 1, 'type' => 0);
        }

        if (strtolower($_cmpFields[0]['field']) == "id")
        {
            if ($_cmpFields[0]['order'] > 0) $this->ksort();
            else $this->krsort();
        }
        else
        {
            // Call uasort() using the static MyCSV::_cmp() function in the class.
            uasort($this->data, array('MyCSV', '_cmp'));
        }
    
        // Reset the internal pointer.
        reset($this->data);
    }

    /**
     * Callback for use in uasort() called in MyCSV::sort().
     *
     * @param a array
     * @param b array
     * @return bool
     * @access private
     */
    function _cmp(&$a, &$b)
    {
        foreach ($GLOBALS['_cmpFields'] as $f)
        {
            // Continue (and maybe return 0) if both arguments are equal.
            if ($a[$f['field']] == $b[$f['field']]) continue;

            // Using this sorting type, empty elements always move to the end.
            if ($f['type'] & SORT_NULL)
                if (! $a[$f['field']] || ! $b[$f['field']]) $f['order'] = -1;

            switch ($f['type'] & ~SORT_NULL)
            {
                case SORT_NUMERIC:
                    // Take both arguments as numbers and return their difference.
                    return ($a[$f['field']] - $b[$f['field']]) * $f['order'];
                case SORT_STRING;
                    // Take both arguments as strings and use strcasecmp() for comparing.
                    return strcasecmp($a[$f['field']], $b[$f['field']]) * $f['order'];
                case SORT_NAT;
                    // Take both arguments as strings and use strcasecmp() for comparing.
                    return strnatcasecmp($a[$f['field']], $b[$f['field']]) * $f['order'];
                default:
                    // By default, thrust in PHP's automatic type conversion.
                    return ($a[$f['field']] > $b[$f['field']] ? 1 : -1) * $f['order'];
            }
        }
        return 0;
    }

    //--------------------------------------------------------------------------
  
    // PHP filesystem functions.
  
    /**
     * Gets a table row including their ID number. Returns false if the row does
     * not exist.
     *
     * @param id int
     * @return array
     */
    function data($id)
    {
        if (! isset($this->data[$id])) return false;
        else return array('id' => $id) + $this->data[$id];
    }

    /**
     * Dumps the table to screen.
     *
     * Example:
     *
     * <pre><?php
     * require_once("MyCSV.class.php");
     * $table = new MyCSV("people");
     * $table->insert(array('name' => "Adam", 'age'  => 23));
     * $table->insert(array('name' => "Bill", 'age'  => 19));
     * echo "&lt;pre>";
     * $table->dump();
     * ?></pre>
     *
     * @return void
     * @see export()
     */
    function dump()
    {
        echo $this->export();
    }

    /**
     * Checks if the CSV file for this table already exists.
     *
     * @return bool
     */
    function exists()
    {
        return file_exists($this->filename);
    }

    /**
     * Returns a complete CSV dump of the table.
     *
     * @return string
     * @see write(), dump()
     */
    function export()
    {
        $csv = implode($this->delimiter, $this->fields) . "\r\n";
        reset($this->data);
        while (list($id, $row) = each($this->data))
        {
            if (strstr($id, '"')) $csv .= '"' . str_replace('"', '""', $id) . '"';
            else $csv .= $id;
            $count = count($this->fields);
            for ($c = 1; $c < $count; $c++)
            {
                if (isset($row[$this->fields[$c]]))
                {
                    // Because fgetcsv() becomes incompatible in PHP 4.3.2.
                    if ($this->_is4_3_2) $row[$this->fields[$c]] = str_replace("\\", "\\\\", $row[$this->fields[$c]]);
                    $csv .= $this->delimiter . '"' . str_replace('"', '""', $row[$this->fields[$c]]) . '"';
                }
                else
                {
                    $csv .= $this->delimiter;
                }
            }
            $csv .= "\r\n";
        }
        return $csv;
    }
  
    /**
     * Checks if the CSV file for this table is writeable.
     *
     * @return bool
     */
    function is_writeable()
    {
      return is_writeable($this->filename);
    }
  
    /**
     * Sets the internal pointer to the data row specified by <i>offset</i>.
     *
     * <i>Whence</i> may be SEEK_SET to set an absolute position (default),
     * SEEK_CUR for a relative position or SEEK_END for an absolute position
     * counted from the end of the table. The behaviour of this method is
     * identical to fseek(). Keep in mind that <i>offset</i> has nothing to do
     * with the row IDs. Example:
     * 
     * <pre>$table = new MyCSV("table");
     * $table->insert(array('id' => 3)); // Row 0
     * $table->insert(array('id' => 7)); // Row 1
     * $table->seek(1);
     * $row = $table->fetch_assoc();
     * echo $row['id']; // Output: 7</pre>
     *
     * @param offset int
     * @param whence int
     * @return bool
     */
    function seek($offset = 0, $whence = SEEK_SET)
    {
        // Calculate absolute offset if end-of-file plus offset is requested.
        if ($whence == SEEK_END) $offset = count($this->data) - 1 - abs($offset);
        // Reset array pointer in SEEK_SET and SEEK_END mode.
        if ($whence != SEEK_CUR) reset($this->data);
        for ($r = 0; $r < $offset; $r++)
        {
            // Return false if offset is out of array bounds.
            if (! next($this->data)) return false;
        }
        return true;
    }
  
    /**
     * Rewrites the CSV file or create a new one. This closes any file when done.
     *
     * @param tablename string
     * @param delimiter string
     * @return bool
     */
    function write($tablename = "", $delimiter = "")
    {
        // Add default file extension if missing.
        if ($tablename && ! preg_match('/\.\w+$/', $tablename)) $tablename .= ".csv";
        // Close the original CSV file and prepare to create a new one.
        if ($tablename && $tablename != $this->filename)
        {
            flock($this->_fp, LOCK_UN);
            fclose($this->_fp);
            $this->_fp = null;
            $this->filename = $tablename;
        }
    
        // Switch to another field delimiter if present.
        if ($delimiter) $this->delimiter = $delimiter;
    
        // Open the CSV file for exclusive writing if not opened already.
        if (! $this->_fp)
        {
            $this->_fp = fopen($this->filename, "wb");
            if (! $this->_fp) return false;
            flock($this->_fp, LOCK_EX);
        }
    
        rewind($this->_fp);
        if (! fwrite($this->_fp, $this->export()))
        {
            // Triggers an user error if the CSV file is write-protected/read-only.
            user_error("MyCSV::write() failed ($this->filename seems to be read only)", E_USER_WARNING);
            return false;
        }
        ftruncate($this->_fp, ftell($this->_fp));
        fflush($this->_fp);
        $this->close();
    
        // Drop empty tables.
        if (count($this->fields) < 2 && ! $this->data)
        {
            unlink($this->filename);
        }
    
        return true;
    }
  
    /**
     * @access private
     */
    function close()
    {
        flock($this->_fp, LOCK_UN);
        fclose($this->_fp);
    }
}

?>